#define _GNU_SOURCE
#include <sched.h>
#include <assert.h>

#include "trx_test.h"
#include "system.h"
#include <semaphore.h>
#include <pthread.h>
#include <math.h>
#include <time.h> /** For SLEEP **/
#include <getopt.h>
#include <stdio.h>

#define NB_ANTENNAS_RX 1
#define NB_ANTENNAS_TX 1

#define SAMPLE_RATE    (122880000)
#define FRAME_SAMPLE_NUM 1228800
//#define TX_RX_ONE_THREAD

#define SLOTS_NUM 10
int txdata_size=61440*SLOTS_NUM;
int rxdata_size=61440*SLOTS_NUM;
int absolute_slot=0;
int nb_slot_frame=2;
int samples_per_subframe=SAMPLE_RATE/1000;

#define TXSEQLEN 61440
int32_t gTxSeq[61440]={0};

//samples_per_subframe 30720, samples_per_slot_wCP 14336.
//txdata_size 57344, txdataF_size 57344, rxdata_size 2465792.

#define VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME(A,B)   

#define NR_TIMESPEC_TO_DOUBLE_US(nr_t)   ( ( (double)nr_t.tv_sec * 1000000 ) + ( (double)nr_t.tv_nsec / 1000 ) )
struct timespec  nr_get_timespec_diff(
       struct timespec *start,
       struct timespec *stop )
{
   struct timespec  result;

 if ( ( stop->tv_nsec - start->tv_nsec ) < 0 ) {
    result.tv_sec = stop->tv_sec - start->tv_sec - 1;
    result.tv_nsec = stop->tv_nsec - start->tv_nsec + 1000000000;
 }
 else {
 result.tv_sec = stop->tv_sec - start->tv_sec;
 result.tv_nsec = stop->tv_nsec - start->tv_nsec;
 }

 return result;
}

openair0_timestamp rx_timestamp, writeTimestamp;
int32_t * rxp[NB_ANTENNAS_RX];
int32_t * txp[NB_ANTENNAS_TX];
int32_t * rxp_temp[NB_ANTENNAS_RX];
int32_t * txp_temp[NB_ANTENNAS_RX];
int readBlockSize, writeBlockSize;
int tx_cnt = 0;
int rx_cnt = 0;
sem_t ric_send_sem;
double clock_gettime_cur_tx;

static void *tx_thread(void *param)
{
  openair0_device *rfdevice = (openair0_device *) param;
  double clock_gettime_cur;
  struct timespec time_start;
  struct timespec time_stop;
  int slot_adv = 2;

  //int mimdelay = 4096*1+2048;
  int mimdelay = 61440;

  while (1) {

    sem_wait (&ric_send_sem);
    //printf("00000000 tx cnt %d rx cnt %d\n", tx_cnt, rx_cnt);
    // use previous timing_advance value to compute writeTimestamp
    //writeTimestamp = timestamp+samples_per_subframe/20*5;// sent after 5 slot.
    writeTimestamp = rx_timestamp+61440+mimdelay;// sent after 5 slot.

    // but use current UE->timing_advance value to compute writeBlockSize
  
    VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_WRITE, VCD_FUNCTION_IN );
    clock_gettime(CLOCK_REALTIME,&time_start);
    #if 1
    
    if ((tx_cnt+slot_adv) % SLOTS_NUM == 5)
    {
      for (int i=0; i<NB_ANTENNAS_TX; i++) {
        memcpy(txp[i] , gTxSeq, TXSEQLEN*4);
      }
    }
    
    for (int i=0; i<NB_ANTENNAS_RX; i++) {
    txp_temp[i] =(int32_t *)txp[i] + ((tx_cnt+slot_adv) % SLOTS_NUM)* 61440;
    
    }

    rfdevice->trx_write_func(rfdevice,
                                              writeTimestamp,
                                               (void **)txp_temp,
                                               writeBlockSize,
                                              1,
                                              1);
    #endif                                          
    clock_gettime(CLOCK_REALTIME,&time_stop ); 
    clock_gettime_cur_tx = NR_TIMESPEC_TO_DOUBLE_US( nr_get_timespec_diff( &time_start, &time_stop ));                                         
    tx_cnt++;
    //printf("11111111 tx cnt %d rx cnt %d\n", tx_cnt, rx_cnt);
    VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_WRITE, VCD_FUNCTION_OUT );

  }

  return NULL;
}

int genSineWave(double freq, double sampleRate, int amp, int len, void *buf)
{
	double dF = 2 * 3.1415926535898 * (freq) / (sampleRate);
	int i;

	int16_t *p16 = buf;

	for (i = 0; i < len; i++)
	{
		p16[i * 2] = cos(dF * i) * amp;
		p16[i * 2 + 1] = sin(dF * i) * amp;
	}

	return 0;
}



int main( int argc, char **argv ) {


 pthread_t tx_threads;

//  test_t *t1=malloc(sizeof(test_t));
  test_t t2[4];
//   t2[0].b = 30720 * 1e3;
//   t1->a = t2[0].b;
// printf("test\n");
// return 0;
  struct sched_param rr_param;
  rr_param.sched_priority = 90;
  sched_setscheduler(getpid(), SCHED_RR, &rr_param);
  cpu_set_t mask;
  CPU_ZERO(&mask);
  CPU_SET(3, &mask);
  int rc = sched_setaffinity(getpid(), sizeof(mask), &mask);
  assert(rc >= 0);

  //T_Config_Init();

  openair0_config_t openair0_cfg[4];
  openair0_device rfdevice;

  int card=0;
  openair0_cfg[card].sample_rate       = 30720 * 1e3 * 4;
  openair0_cfg[card].samples_per_frame = 30720 * 4;

  openair0_cfg[card].duplex_mode = duplex_mode_TDD;

  openair0_cfg[card].Mod_id = 0;
  openair0_cfg[card].num_rb_dl = 51;
  openair0_cfg[card].clock_source = 0;
  openair0_cfg[card].time_source = 0;
  openair0_cfg[card].tx_num_channels = NB_ANTENNAS_TX;
  openair0_cfg[card].rx_num_channels = NB_ANTENNAS_RX;
  openair0_cfg[card].rx_gain[0] = 80;
  openair0_cfg[card].tx_gain[0] = 90;

  printf("HW: Configuring card %d, sample_rate %f, tx/rx num_channels %d/%d, duplex_mode %d\n",
      card,
      openair0_cfg[card].sample_rate,
      openair0_cfg[card].tx_num_channels,
      openair0_cfg[card].rx_num_channels,
      openair0_cfg[card].duplex_mode);

  double dl_carrier = 3349380000;
  double ul_carrier = 3349380000;
  int freq_off   = 0;
  int rx_gain_off = 0;
  int rf_chain = 0;
  double freq_scale  = (double)(dl_carrier + freq_off) / dl_carrier;

  for (int i = rf_chain; i < rf_chain + 1; i++) {
    printf("test\n");
    if (i < openair0_cfg->rx_num_channels)
      openair0_cfg->rx_freq[i + rf_chain] = dl_carrier * freq_scale;
    else
      openair0_cfg->rx_freq[i] = 0.0;

    if (i<openair0_cfg->tx_num_channels)
      openair0_cfg->tx_freq[i] = ul_carrier * freq_scale;
    else
      openair0_cfg->tx_freq[i] = 0.0;

    openair0_cfg->autocal[i] = 1;

    //if (i < openair0_cfg->rx_num_channels) 
    {
      printf("HW: Configuring channel %d (rf_chain %d): setting tx_freq %f Hz, rx_freq %f Hz\n",
        i,
        rf_chain,
        openair0_cfg->tx_freq[i],
        openair0_cfg->rx_freq[i]);
    }
  }
//    nr_rf_card_config_gain(&openair0_cfg[card], rx_gain_off);

  openair0_cfg[card].sdr_addrs = "dev=pciex:0,auxdac1=300";

  openair0_timestamp timestampold ;

  int start_rx_stream = 1;

  double clock_gettime_cur;
  struct timespec time_start, time_start1;
  struct timespec time_stop, time_stop1;

  int res = sem_init(&ric_send_sem, 0, 0);  
  if (res != 0)  
  {  
  perror("Semaphore initialization failed");  
  }  



  for (int i=0; i<NB_ANTENNAS_RX; i++) {
    rxp[i] = (int32_t *) malloc16_clear( rxdata_size * sizeof(int32_t));
  }
  for (int i=0; i<NB_ANTENNAS_TX; i++) {
    txp[i] = (int32_t *) malloc16_clear( txdata_size * sizeof(int32_t));
    txp_temp[i] = (int32_t *) malloc16_clear( 61440 * sizeof(int32_t));
  }

  

  printf("sine wave\n");
	
  genSineWave(3.84e6, 122.88e6, 0x7fff, TXSEQLEN, gTxSeq);

  readBlockSize = samples_per_subframe/2;
  writeBlockSize = readBlockSize;
    
  device_init(&(rfdevice), &openair0_cfg[0]);


  threadCreate(&tx_threads, tx_thread, &rfdevice, "tx_thread", 5, sched_get_priority_max(SCHED_RR)-2);
  rfdevice.trx_start_func(&rfdevice);

//first time is very long, don't know why.
  clock_gettime(CLOCK_REALTIME,&time_start);
  VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_READ, VCD_FUNCTION_IN );
  rfdevice.trx_read_func(&rfdevice,
                                          &rx_timestamp,
                                          (void **)rxp,
                                          readBlockSize,
                                          1);
  timestampold = rx_timestamp;
  VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_READ, VCD_FUNCTION_OUT );
  clock_gettime(CLOCK_REALTIME,&time_stop);
  clock_gettime_cur = NR_TIMESPEC_TO_DOUBLE_US( nr_get_timespec_diff( &time_start, &time_stop ));//us
  printf("using %.2lf msec in 1st time\n", clock_gettime_cur/1000);
  clock_gettime(CLOCK_REALTIME,&time_start);

  
  while (1) {

     clock_gettime(CLOCK_REALTIME,&time_stop);

    clock_gettime_cur = NR_TIMESPEC_TO_DOUBLE_US( nr_get_timespec_diff( &time_start, &time_stop ));//us
    if ((clock_gettime_cur > 510) || (clock_gettime_cur_tx > 510))
         printf("slot_num %d, timestamp %ld, writeTimestamp %ld,time = %.2f sec, delay_time=(%.2lf, %d) (tx %.2lf, %d)msec. readBlockSize %d. real read %ld.\n",
          absolute_slot, rx_timestamp,  writeTimestamp, ((float)absolute_slot)/2/1000, clock_gettime_cur/1000,rx_cnt, clock_gettime_cur_tx/1000, tx_cnt, readBlockSize,(rx_timestamp - timestampold));
    clock_gettime(CLOCK_REALTIME,&time_start);
    timestampold = rx_timestamp;

#if 1
    VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_READ, VCD_FUNCTION_IN );
    for (int i=0; i<NB_ANTENNAS_RX; i++) {
    rxp_temp[i] =(int32_t *)rxp[i] + (rx_cnt % SLOTS_NUM)* 61440;
    }
    rfdevice.trx_read_func(&rfdevice,
                                           &rx_timestamp,
                                           //(void **)&rxp[0][(rx_cnt % SLOTS_NUM)* 61440],
                                           (void **)rxp_temp,
                                           readBlockSize,
                                           1);
    rx_cnt++;
    absolute_slot++;

 
    VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_READ, VCD_FUNCTION_OUT );
#endif
#if 0
    if(rx_cnt == 2000)
    {
      printf("begin to dump\n");
			FILE *fp1 = fopen("rfdata_v3_ant0.bin", "wb");
			fwrite(&rxp[0][0], 4, rxdata_size, fp1);
			fclose(fp1);

			printf("end, write file ok ++++++++++++++++++++++++\r\n");
			usleep(50000);
			exit(0);
    }
#endif
#ifdef TX_RX_ONE_THREAD
    // use previous timing_advance value to compute writeTimestamp
    //writeTimestamp = timestamp+samples_per_subframe/20*5;// sent after 5 slot.
    writeTimestamp = rx_timestamp+samples_per_subframe/4;// sent after 5 slot.

    // but use current UE->timing_advance value to compute writeBlockSize
  
    VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_WRITE, VCD_FUNCTION_IN );
    clock_gettime(CLOCK_REALTIME,&time_start1 ); 
    rfdevice.trx_write_func(&rfdevice,
                                              writeTimestamp,
                                              (void **)txp,
                                              writeBlockSize,
                                              1,
                                              1);
    clock_gettime(CLOCK_REALTIME,&time_stop1 ); 
    clock_gettime_cur_tx = NR_TIMESPEC_TO_DOUBLE_US( nr_get_timespec_diff( &time_start1, &time_stop1 ));  
    VCD_SIGNAL_DUMPER_DUMP_FUNCTION_BY_NAME( VCD_SIGNAL_DUMPER_FUNCTIONS_TRX_WRITE, VCD_FUNCTION_OUT );
#else
    sem_post(&ric_send_sem); 
#endif  
} // while !oai_exit

  return 0;
}
